'''
    Return x, y, and metadata for cursor position or multiple points in selected box.
    Values are formatted into Panel StaticText and put in row/col format (cursor location)
    or single line (box location).
'''

import numpy as np
from pandas import to_datetime
import panel as pn

from vidavis.plot.ms_plot._ms_plot_constants import TIME_FORMAT

def cursor_changed(x, y, last_cursor):
    ''' Check whether cursor position changed '''
    if not x and not y:
        return False # not cursor callback
    if last_cursor and last_cursor == (x, y):
        return False # same cursor
    return True # new cursor or cursor changed

def points_changed(data, last_points):
    ''' Check whether point positions changed '''
    # No data = {'x': [], 'y': []}
    if len(data['x']) == 0 and len(data['y']) == 0:
        return False # not points callback
    if last_points and last_points == data:
        return False # same points
    return True # new points, points changed, or points deleted

def box_changed(bounds, last_box):
    ''' Check whether box position changed '''
    # No bounds = None
    if not bounds:
        return False # no data, not box select callback
    if last_box and last_box == bounds:
        return False # same box
    return True # new box, box changed, or box deleted

def locate_point(xds, position, vis_axis):
    '''
        Get cursor location as values of coordinates and data vars.
            xds (Xarray Dataset): data for plot
            position (dict): {coordinate: value} of x and y axis positions
            vis_axis (str): visibility component of complex value
        Returns:
            list of pn.widgets.StaticText(name, value) with value formatted for its type
    '''
    static_text_list = []
    values, units = _get_point_location(xds, position, vis_axis)

    # Rename baseline coordinate names to not confuse user for selection
    baseline_names = {'baseline': 'baseline_index', 'baseline_name': 'baseline', 'antenna_name': 'antenna_index', 'antenna': 'antenna_name'}
    for name, value in values.items():
        name = baseline_names[name] if name in baseline_names else name
        static_text = _get_location_text(name, value, units)
        static_text_list.append(static_text)
    return static_text_list

def locate_box(xds, bounds, vis_axis):
    '''
        Get location of each point in box bounds as values of coordinate and data vars.
            xds (Xarray Dataset): data for plot
            bounds (dict): {coordinate: (start, end)} of x and y axis ranges
            vis_axis (str): visibility component of complex value
        Returns:
            list of list of pn.widgets.StaticText(name, value), one list per point.
    '''
    points = []
    npoints = 0

    if xds:
        try:
            selection = {}
            for coord, val in bounds.items():
                # Round index values to int for selection
                selection[coord] = slice(_get_selection_value(coord, val[0]), _get_selection_value(coord, val[1]))
            sel_xds = xds.sel(indexers=None, method=None, tolerance=None, drop=False, **selection)

            x_coord, y_coord = bounds.keys()
            npoints = sel_xds.sizes[x_coord] * sel_xds.sizes[y_coord]
            counter = 0

            for y in sel_xds[y_coord].values:
                for x in sel_xds[x_coord].values:
                    position = {x_coord: x, y_coord: y}
                    points.append(locate_point(sel_xds, position, vis_axis))
                    counter += 1
                    if counter == 100:
                        break
                if counter == 100:
                    break
        except KeyError:
            pass
    return npoints, points

def _get_point_location(xds, position, vis_axis):
    ''' Select plot data xds with point x, y position, and return coord and data_var values describing the location.
            xds (Xarray Dataset): data for plot
            position (dict): {coordinate: value} of x and y axis positions
            vis_axis (str): visibility component of complex value
        Returns:
            values (dict): {name: value} for each location item
            units (dict): {name: unit} for each value which has a unit defined.
    '''
    values = position.copy()
    units = {}

    if xds:
        try:
            for coord, value in position.items():
                # Round index coordinates to int and convert time to datetime if float for selection
                position[coord] = _get_selection_value(coord, value)

            sel_xds = xds.sel(indexers=None, method='nearest', tolerance=None, drop=False, **position)
            for coord in sel_xds.coords:
                if coord == 'uvw_label' or ('baseline_antenna' in coord and 'baseline_name' in sel_xds.coords):
                    continue
                val, unit = _get_xda_val_unit(sel_xds[coord])
                values[coord] = val
                units[coord] = unit
            for data_var in sel_xds.data_vars:
                if 'TIME_CENTROID' in data_var:
                    continue
                val, unit = _get_xda_val_unit(sel_xds[data_var])
                if data_var == 'UVW':
                    names = ['U', 'V', 'W']
                    for i, name in enumerate(names):
                        values[name] = val[i]
                        units[name] = unit
                else:
                    values[data_var] = val
                    units[data_var] = unit
        except KeyError:
            pass

    # Set complex component name for visibilities
    if 'VISIBILITY' in values:
        values[vis_axis.upper()] = values.pop('VISIBILITY')
    return values, units

def _get_selection_value(coord, value):
    ''' Convert index coordinates to int and float time coordinate to datetime '''
    if coord in ['baseline', 'antenna_name', 'polarization']:
        # Round index coordinates to int for selecction
        value = round(value)
    elif coord == 'time' and isinstance(value, float):
        # Bokeh datetime values are floating-point numbers: milliseconds since the Unix epoch
        value = to_datetime(value, unit='ms', origin='unix')
    return value

def _get_xda_val_unit(xda):
    ''' Return value and unit of xda (selected so only one value) '''
    # Value
    value = xda.values
    if isinstance(value, np.ndarray) and value.size == 1:
        value = value.item()

    # Unit
    try:
        unit = xda.attrs['units']
        unit = unit[0] if (isinstance(unit, list) and len(unit) == 1) else unit
        unit = '' if unit == 'unkown' else unit
    except KeyError:
        unit = ''

    return value, unit

def _get_location_text(name, value, units):
    ''' Format value and unit (if any) and return Panel StaticText '''
    if not isinstance(value, str):
        # Format numeric and datetime values
        if name == "FLAG":
            value = "nan" if np.isnan(value) else int(value)
        elif isinstance(value, float):
            if np.isnan(value):
                value = "nan"
            elif value < 1e6:
                value = f"{value:.4f}"
            else:
                value = f"{value:.4e}"
        elif isinstance(value, np.datetime64):
            value = to_datetime(np.datetime_as_string(value)).strftime(TIME_FORMAT)
            units.pop(name) # no unit for datetime string
    unit = units[name] if name in units else ""
    return pn.widgets.StaticText(name=name, value=f"{value} {unit}")



''' Get MeasurementSet data from xarray Dataset '''

from astropy import constants
import numpy as np
import xarray as xr

from vidavis.plot.ms_plot._ms_plot_constants import SPECTRUM_AXIS_OPTIONS, UVW_AXIS_OPTIONS, VIS_AXIS_OPTIONS, WEIGHT_AXIS_OPTIONS

def get_correlated_data(xds, data_group):
    ''' Return correlated_data value in data_group dict '''
    return xds.attrs['data_groups'][data_group]['correlated_data']

def get_axis_data(xds, axis, data_group=None):
    ''' Get requested axis data from xarray dataset.
            xds (dict): msv4 xarray.Dataset
            axis (str): axis data to retrieve.
            data_group (str): correlated data group name.
        Returns:    xarray.DataArray
    '''
    group_info = xds.data_groups[data_group]
    xda = None
    if _is_coordinate_axis(axis):
        xda = xds[axis]
    elif _is_antenna_axis(axis):
        xda = _get_antenna_axis(xds, axis)
    elif _is_vis_axis(axis):
        xda = _calc_vis_axis(xds, axis, group_info)
    elif _is_uvw_axis(axis):
        xda = _calc_uvw_axis(xds, axis, group_info)
    elif _is_weight_axis(axis):
        xda = _calc_weight_axis(xds, axis, group_info)
    elif 'spw' in axis:
        xda = _get_spw_axis(xds, axis)
    elif 'wave' in axis:
        xda = _calc_wave_axis(xds, axis, group_info)
    elif axis == 'field':
        xda = xr.DataArray([xds[group_info['correlated_data']].field_and_source_xds.field_name])
    elif axis == 'flag':
        xda = xds[group_info['flag']]
    elif axis == 'intents':
        xda = xr.DataArray(["".join(xds.partition_info['intents'])])
    elif axis == 'channel':
        xda = xr.DataArray(np.array(range(xds.frequency.size)))
    return xda

def _is_coordinate_axis(axis):
    return axis in ['scan_number', 'time', 'frequency', 'polarization',
        #'velocity': 'frequency', # calculate
        # TODO?
        #'observation':  no id, xds.observation_info (observer, project, release date)
        #'feed1':  no id, xds.antenna_xds
        #'feed2':  no id, xds.antenna_xds
    ]

def _is_vis_axis(axis):
    return axis in VIS_AXIS_OPTIONS

def _is_uvw_axis(axis):
    return axis in UVW_AXIS_OPTIONS

def _is_antenna_axis(axis):
    return 'baseline' in axis or 'antenna' in axis

def _is_weight_axis(axis):
    return axis in WEIGHT_AXIS_OPTIONS

def _get_spw_axis(xds, axis):
    if axis == 'spw_id':
        return xr.DataArray([xds.frequency.spectral_window_id])
    if axis == 'spw_name':
        return xr.DataArray([xds.frequency.spectral_window_name])
    raise ValueError(f"Invalid spw axis {axis}")

def _calc_vis_axis(xds, axis, group_info):
    ''' Calculate axis from correlated data '''
    correlated_data = group_info['correlated_data']
    xda = xds[correlated_data]

    # Single dish spectrum
    if correlated_data == "SPECTRUM":
        if axis in SPECTRUM_AXIS_OPTIONS:
            return xda.assign_attrs(units='Jy')
        raise RuntimeError(f"Vis axis {axis} invalid for SPECTRUM dataset, select from {SPECTRUM_AXIS_OPTIONS}")

    # Interferometry visibilities
    if axis == 'amp':
        return np.absolute(xda).assign_attrs(units='Jy')
    if axis == 'phase':
        # np.angle(xda) returns ndarray not xr.DataArray
        return (np.arctan2(xda.imag, xda.real) * 180.0/np.pi).assign_attrs(units="deg")
    if axis == 'real':
        return np.real(xda.assign_attrs(units='Jy'))
    if axis == 'imag':
        return np.imag(xda.assign_attrs(units='Jy'))
    return None

def _calc_uvw_axis(xds, axis, group_info):
    ''' Calculate axis from UVW xarray DataArray '''
    if 'uvw' not in group_info:
        raise RuntimeError(f"Axis {axis} is not valid in this dataset, no uvw data")

    uvw_xda = xds[group_info['uvw']]

    if axis == 'u':
        return uvw_xda.isel(uvw_label=0)
    if axis == 'v':
        return uvw_xda.isel(uvw_label=1)
    if axis == 'w':
        return uvw_xda.isel(uvw_label=2)

    # uvdist
    u_xda = uvw_xda.isel(uvw_label=0)
    v_xda = uvw_xda.isel(uvw_label=1)
    return np.sqrt(np.square(u_xda) + np.square(v_xda))

def _calc_wave_axis(xds, axis, group_info):
    wave_axes = {'uwave': 'u', 'vwave': 'v', 'wwave': 'w', 'uvwave': 'uvdist'}
    if axis not in wave_axes:
        raise ValueError(f"Invalid wave axis {axis}")
    uvwdist_array = _calc_uvw_axis(xds, wave_axes[axis], group_info).values / constants.c
    uvwdist_len = len(uvwdist_array)
    freq_array = xds.frequency.values
    wave = np.zeros(shape=(len(freq_array), uvwdist_len), dtype=np.double)

    for i, in range(uvwdist_len):
        wave[:, i] = uvwdist_array[i] * freq_array

    wave_xda = xr.DataArray(wave, dims=['frequency', 'uvw_label'])
    return wave_xda

def _calc_weight_axis(xds, axis, group_info):
    weight = xds[group_info['weight']]
    if axis == 'weight':
        return weight
    return np.sqrt(1.0 / weight)

def _get_antenna_axis(xds, axis):
    if 'antenna_name' in xds.coords:
        return xds.antenna_name

    if axis == 'antenna1':
        return xds.baseline_antenna1_name
    if axis == 'antenna2':
        return xds.baseline_antenna2_name
    if axis == 'baseline_id':
        return xds.baseline_id
    if axis == 'baseline':
        return xds.baseline
    raise ValueError(f"Invalid antenna/baseline axis {axis}")



'''
    Class for accessing and selecting MeasurementSet data using a data backend.
'''

from vidavis.data.measurement_set.processing_set._ps_data import PsData

#pylint: disable=too-many-public-methods
class MsData:
    '''
    Access and select MeasurementSet data.
    Current backend implementation is PsData using xradio Processing Set.
    '''

    def __init__(self, ms_path, logger):
        self._ms_path = ms_path
        self._logger = logger
        self._data = None
        self._data_initialized = False
        self._init_data(ms_path)

    def is_valid(self):
        ''' Returns whether MS path has been set so data can be accessed. '''
        return self._data_initialized

    def is_ms_path(self, path):
        ''' Check if input path matches input ms path or zarr path '''
        return path == self.get_path() or path == self._ms_path

    def get_path(self):
        ''' Returns path of zarr file, MS file, or None if not set. '''
        if self._data_initialized:
            return self._data.get_path() # path to zarr file
        return self._ms_path # path to ms v2

    def summary(self, data_group='base', columns=None):
        ''' Print summary of Processing Set data.
                columns (None, str, list): type of metadata to list.
                    None:      Print all summary columns in ProcessingSet.
                    'by_msv4': Print formatted summary metadata by MSv4.
                    str, list: Print a subset of summary columns in ProcessingSet.
                        Options: 'name', 'intents', 'shape', 'polarization', 'scan_number', 'spw_name',
                                 'field_name, 'source_name', 'field_coords', 'start_frequency', 'end_frequency'
        '''
        # ProcessingSet function
        if self._data_initialized:
            self._data.summary(data_group, columns)
        else:
            self._log_no_ms()

    def get_summary(self):
        ''' Return Pandas DataFrame summary of ProcessingSet '''
        if self._data_initialized:
            return self._data.get_summary()
        self._log_no_ms()
        return None

    def data_groups(self):
        ''' Returns set of data group names in Processing Set data. '''
        if self._data_initialized:
            return self._data.get_data_groups()
        self._log_no_ms()
        return None

    def plot_antennas(self, label_antennas=False):
        ''' Plot antenna positions.
                label_antennas (bool): label positions with antenna names.
        '''
        if self._data_initialized:
            self._data.plot_antennas(label_antennas)
        else:
            self._log_no_ms()

    def plot_phase_centers(self, data_group='base', label_all_fields=False):
        ''' Plot the phase center locations of all fields in the Processing Set (original or selected) and label central field.
                label_all_fields (bool); label all fields on the plot
                data_group (str); data group to use for processing.
        '''
        if self._data_initialized:
            self._data.plot_phase_centers(label_all_fields, data_group)
        else:
            self._log_no_ms()

    def get_num_ms(self):
        ''' Returns number of MeasurementSets in data. '''
        if self._data_initialized:
            return self._data.get_ps_len()
        self._log_no_ms()
        return None

    def get_max_data_dims(self):
        ''' Returns maximum length of dimensions in data. '''
        if self._data_initialized:
            return self._data.get_max_dims()
        self._log_no_ms()
        return None

    def get_data_dimensions(self):
        ''' Returns names of data dimensions. '''
        if self._data_initialized:
            return self._data.get_data_dimensions()
        self._log_no_ms()
        return None

    def get_dimension_values(self, dim):
        ''' Returns values for dimension in current data.
                dim (str): dimension name
        '''
        if self._data_initialized:
            return self._data.get_dimension_values(dim)
        self._log_no_ms()
        return None

    def get_dimension_attrs(self, dim):
        ''' Returns dict of data attributes for dimension.
                dim (str): dimension name
        '''
        if self._data_initialized:
            return self._data.get_dimension_attrs(dim)
        self._log_no_ms()
        return None

    def get_first_spw(self, data_group='base'):
        ''' Returns name of first spw by id. '''
        if self._data_initialized:
            return self._data.get_first_spw(data_group)
        self._log_no_ms()
        return None

    def select_ps(self, query=None, string_exact_match=True, **kwargs):
        ''' Apply data group and summary column selection to ProcessingSet. See ProcessingSetXdt query().
            https://xradio.readthedocs.io/en/latest/measurement_set/schema_and_api/measurement_set_api.html#xradio.measurement_set.ProcessingSetXdt.query
            Selections are cumulative until clear_selection() is called.
            Saves selected ProcessingSet internally.
            Throws exception if selection fails.
        '''
        if self._data_initialized:
            self._data.select_ps(query=query, string_exact_match=string_exact_match, **kwargs)
        else:
            self._log_no_ms()

    def select_ms(self, indexers=None, method=None, tolerance=None, drop=False, **indexers_kwargs):
        ''' Apply dimension and data group selection to MeasurementSet. See MeasurementsSetXdt sel().
            https://xradio.readthedocs.io/en/latest/measurement_set/schema_and_api/measurement_set_api.html#xradio.measurement_set.MeasurementSetXdt.sel.
            Additional supported selection besides dimensions include "baseline", "antenna1", "antenna2".
            Selections are cumulative until clear_selection() is called.
            Saves selected ProcessingSet internally.
            Throws exception if selection fails.
        '''
        if self._data_initialized:
            self._data.select_ms(indexers, method, tolerance, drop, **indexers_kwargs)
        else:
            self._log_no_ms()

    def clear_selection(self):
        ''' Clears selection dict and selected data. '''
        if self._data_initialized:
            self._data.clear_selection()

    def get_vis_stats(self, selection, vis_axis):
        ''' Returns statistics (min, max, mean, std) for data in data group selected by selection.
                selection (dict): fields and values to select
                vis_axis (str): complex component to apply to data
        '''
        if self._data_initialized:
            return self._data.get_vis_stats(selection, vis_axis)
        self._log_no_ms()
        return None

    def get_correlated_data(self, data_group):
        ''' Returns name of correlated data variable in Processing Set data group '''
        if self._data_initialized:
            return self._data.get_correlated_data(data_group)
        self._log_no_ms()
        return None

    def get_raster_data(self, plot_inputs):
        ''' Returns xarray Dataset after applying plot inputs and raster plane selection '''
        if self._data_initialized:
            return self._data.get_raster_data(plot_inputs)
        self._log_no_ms()
        return None

    def _log_no_ms(self):
        ''' Standardized log message when path has not been set. '''
        self._logger.info("No MS path set, cannot access data")

    def _init_data(self, ms_path):
        ''' Data backend for MeasurementSet; currently xradio ProcessingSet '''
        if ms_path:
            self._data = PsData(ms_path, self._logger)
            self._data_initialized = True
#pylint: enable=too-many-public-methods



'''
MeasurementSet data backend using xradio Processing Set.
'''

import numpy as np
import pandas as pd

try:
    from vidavis.data.measurement_set.processing_set._ps_io import get_processing_set
    _HAVE_XRADIO = True
    from vidavis.plot.ms_plot._ms_plot_constants import TIME_FORMAT
    from vidavis.data.measurement_set.processing_set._ps_select import select_ps, select_ms
    from vidavis.data.measurement_set.processing_set._ps_stats import calculate_ps_stats
    from vidavis.data.measurement_set.processing_set._ps_raster_data import raster_data
    from vidavis.data.measurement_set.processing_set._xds_data import get_correlated_data
except ImportError as e:
    _HAVE_XRADIO = False


class PsData:
    '''
    Class implementing data backend using xradio Processing Set for accessing and selecting MeasurementSet data.
    '''

    def __init__(self, ms, logger):
        if not _HAVE_XRADIO:
            raise RuntimeError("xradio package not available for reading MeasurementSet")

        if not ms:
            raise RuntimeError("MS path not available for reading MeasurementSet")

        # Open processing set from zarr. Converts msv2 if ms path is not zarr
        self._ps_xdt, self._zarr_path = get_processing_set(ms, logger)

        self._logger = logger
        self._selection = {}
        self._selected_ps_xdt = None # cumulative selection

    def get_path(self):
        ''' Return path to zarr file (input or converted from msv2) '''
        return self._zarr_path

    def summary(self, data_group='base', columns=None):
        ''' Print full or selected summary of Processing Set metadata, optionally by ms '''
        ps_summary = self.get_summary(data_group)
        pd.set_option("display.max_rows", len(self._ps_xdt))
        pd.set_option("display.max_columns", len(ps_summary.columns))
        pd.set_option("display.max_colwidth", None)

        if columns is None:
            print(ps_summary)
        elif columns == "by_ms":
            for row in ps_summary.itertuples(index=False):
                print(f"MSv4 name: {row[0]}")
                print(f"intent: {row[1]}")
                shape = row[2]
                print(f"shape: {shape[0]} times, {shape[1]} baselines, {shape[2]} channels, {shape[3]} polarizations")
                print(f"polarization: {row[3]}")
                scans = [str(scan) for scan in row[4]]
                print(f"scan_name: {scans}")
                print(f"spw_name: {row[5]}")
                fields = [str(field) for field in row[6]]
                print(f"field_name: {fields}")
                sources = [str(source) for source in row[7]]
                print(f"source_name: {sources}")
                lines = [str(line) for line in row[8]]
                print(f"line_name: {lines}")
                field_coords = row[9]
                print(f"field_coords: ({field_coords[0]}) {field_coords[1]} {field_coords[2]}")
                print(f"frequency range: {row[10]:e} - {row[11]:e}")
                print("-----")
        else:
            if isinstance(columns, str):
                columns = [columns]
            col_df = ps_summary[columns]
            print(col_df)

    def get_summary(self, data_group='base'):
        ''' Return summary of original ps '''
        return self._ps_xdt.xr_ps.summary(data_group)

    def get_data_groups(self):
        ''' Returns set of data group names in Processing Set data. '''
        data_groups = []
        for ms_xdt_name in self._ps_xdt:
            data_groups.extend(list(self._ps_xdt[ms_xdt_name].data_groups))
        return set(data_groups)

    def plot_antennas(self, label_antennas=False):
        ''' Plot antenna positions.
                label_antennas (bool): label positions with antenna names.
        '''
        self._ps_xdt.xr_ps.plot_antenna_positions(label_antennas)

    def plot_phase_centers(self, label_all_fields=False, data_group='base'):
        ''' Plot the phase center locations of all fields in the Processing Set (original or selected) and label central field.
                label_all_fields (bool); label all fields on the plot
                data_group (str); data group to use for processing.
        '''
        self._ps_xdt.xr_ps.plot_phase_centers(label_all_fields, data_group)

    def get_ps_len(self):
        ''' Returns number of ms_xdt in selected ps_xdt (if selected) '''
        return len(self._get_ps_xdt())

    def get_max_dims(self):
        ''' Returns maximum length of dimensions in selected ProcessingSet (if selected) '''
        ps_xdt = self._get_ps_xdt()
        return ps_xdt.xr_ps.get_max_dims()

    def get_data_dimensions(self):
        ''' Return names of the data dimensions. '''
        dims = list(self.get_max_dims().keys())
        if 'uvw_label' in dims:
            dims.remove('uvw_label') # not a VISIBILITY/SPECTRUM data dim
        dims = ['baseline' if dim=='baseline_id' else dim for dim in dims]
        return dims

    def get_dimension_values(self, dimension):
        ''' Return sorted list of unique values for input dimension in selected ProcessingSet.
            For 'time', returns datetime strings.
            For spectrum datasets, 'antenna2' returns empty list.
        '''
        ps_xdt = self._get_ps_xdt()
        dim_values = []

        if dimension == 'time':
            dim_values = self._get_time_strings(ps_xdt)
        elif dimension == 'baseline':
            dim_values = self._get_baselines(ps_xdt)
        else:
            if dimension == 'antenna1':
                dimension = 'baseline_antenna1_name'
            elif dimension == 'antenna2':
                dimension = 'baseline_antenna2_name'
            for ms_xdt in ps_xdt.values():
                if dimension not in ms_xdt.coords:
                    if 'antenna1' in dimension and 'antenna_name' in ms_xdt.coords:
                        dimension = 'antenna_name' # spectrum dataset
                    else:
                        continue
                try:
                    dim_values.extend([value.item() for value in ms_xdt[dimension].values])
                except TypeError:
                    dim_values.append(ms_xdt[dimension].values.item())

        if not dim_values:
            return dim_values
        return sorted(set(dim_values))

    def _get_time_strings(self, ps):
        ''' Return time values as string not float '''
        times = []
        for ms_xdt in ps.values():
            time_xda = ms_xdt.time
            time_attrs = time_xda.attrs
            date_strings = pd.to_datetime(time_xda, unit=time_attrs['units'][0], origin=time_attrs['format']).strftime(TIME_FORMAT).values
            times.extend(date_strings.tolist())
        return times

    def _get_baselines(self, ps):
        ''' Return baseline strings as ant1_name & ant2_name.
            For spectrum datasets, return list of antenna_name. '''
        baselines = []
        for ms_xdt in ps.values():
            if 'antenna_name' in ms_xdt.coords:
                baselines.extend(ms_xdt.antenna_name.values)
            else:
                ant1_names = ms_xdt.baseline_antenna1_name.values
                ant2_names = ms_xdt.baseline_antenna2_name.values
                for ant1, ant2 in zip(ant1_names, ant2_names):
                    baselines.append(f"{ant1} & {ant2}")
        return baselines

    def get_dimension_attrs(self, dim):
        ''' Return attributes dict for input dimension in ProcessingSet. '''
        ps_xdt = self._get_ps_xdt()
        return ps_xdt.get(0)[dim].attrs

    def get_first_spw(self, data_group='base'):
        ''' Return first spw name in selected ps summary '''
        spw_names = self.get_summary(data_group)['spw_name']
        return spw_names[0]

    def select_ps(self, query=None, string_exact_match=True, **kwargs):
        ''' Apply data group and summary column selection to ProcessingSet. See ProcessingSetXdt query().
            https://xradio.readthedocs.io/en/latest/measurement_set/schema_and_api/measurement_set_api.html#xradio.measurement_set.ProcessingSetXdt.query
            Also applies selection to ms_xdt in ps.
            Selections are cumulative until clear_selection() is called.
            Saves selected ProcessingSet internally.
            Throws exception if selection fails.
        '''
        ps_xdt = self._get_ps_xdt()
        self._selected_ps_xdt = select_ps(ps_xdt, self._logger, query=query, string_exact_match=string_exact_match, **kwargs)

    def select_ms(self, indexers=None, method=None, tolerance=None, drop=False, **indexers_kwargs):
        ''' Apply dimension and data group selection to MeasurementSet. See MeasurementsSetXdt sel().
            https://xradio.readthedocs.io/en/latest/measurement_set/schema_and_api/measurement_set_api.html#xradio.measurement_set.MeasurementSetXdt.sel.
            Additional supported selection besides dimensions include "baseline", "antenna1", "antenna2".
            Selections are cumulative until clear_selection() is called.
            Saves selected ProcessingSet internally.
            Throws exception if selection fails.
        '''
        ps_xdt = self._get_ps_xdt()
        self._selected_ps_xdt = select_ms(ps_xdt, self._logger, indexers, method, tolerance, drop, **indexers_kwargs)

    def clear_selection(self):
        ''' Clear previous selections and use original ps_xdt '''
        self._selected_ps_xdt = None

    def get_vis_stats(self, ps_selection, vis_axis):
        ''' Returns statistics (min, max, mean, std) for data in data group selected by selection.
                data_group (str): correlated data to use for calculations
                selection (dict): fields and values to select
                vis_axis (str): complex component to apply to data
        '''
        stats_ps_xdt = select_ps(self._ps_xdt, self._logger, query=None, string_exact_match=True, **ps_selection)
        data_group = ps_selection['data_group_name'] if 'data_group_name' in ps_selection else 'base'
        return calculate_ps_stats(stats_ps_xdt, self._zarr_path, vis_axis, data_group, self._logger)

    def get_correlated_data(self, data_group):
        ''' Returns name of 'correlated_data' in Processing Set data_group '''
        ps_xdt = self._get_ps_xdt()
        for ms_xdt in ps_xdt.values():
            if data_group in ms_xdt.attrs['data_groups']:
                return get_correlated_data(ms_xdt.ds, data_group)
        raise RuntimeError(f"No correlated data for data group {data_group}")

    def get_raster_data(self, plot_inputs):
        ''' Returns xarray Dataset after applying plot inputs and raster plane selection '''
        return raster_data(self._get_ps_xdt(),
            plot_inputs,
            self._logger
        )

    def _get_ps_xdt(self):
        ''' Returns selected ps_xdt if selection has been done, else original ps_xdt '''
        return self._selected_ps_xdt if self._selected_ps_xdt else self._ps_xdt

    def _get_unique_values(self, df_col):
        ''' Return unique values in pandas Dataframe column, for summary '''
        values = df_col.to_numpy()
        try:
            # numeric arrays
            return np.unique(np.concatenate(values))
        except ValueError:
            # string arrays
            all_values = [row[0] for row in values]
            return np.unique(np.concatenate(all_values))



''' Utilities for inputs to MeasurementSet plots '''

def inputs_changed(plot_inputs, last_plot_inputs):
    ''' Check if inputs changed and need new plot '''
    if not last_plot_inputs:
        return True

    for key, val in plot_inputs.items():
        if not _values_equal(val, last_plot_inputs[key]):
            return True
    return False

def _values_equal(val1, val2):
    ''' Test if values are set and equal, or not set (cannot compare value with None) '''
    if val1 is not None and val2 is not None: # both set
        return val1 == val2
    if val1 is None and val2 is None: # both None
        return True
    return False # one set and other is None



'''
Check inputs to MsRaster plot() or its GUI
'''

from vidavis.plot.ms_plot._ms_plot_constants import VIS_AXIS_OPTIONS, AGGREGATOR_OPTIONS

def check_inputs(inputs):
    ''' Check plot input types, and axis (plot, agg, iter) input values. '''
    _set_baseline_antenna_axis(inputs)
    _check_axis_inputs(inputs)
    _check_agg_inputs(inputs)
    _check_color_inputs(inputs)
    _check_other_inputs(inputs)

def _set_baseline_antenna_axis(inputs):
    ''' Set baseline axis to dimension in data_dims '''
    if 'data_dims' not in inputs:
        return

    data_dims = inputs['data_dims']
    baseline_dim = 'antenna_name' if 'antenna_name' in data_dims else 'baseline'

    # Convert baseline axis to existing baseline dimension
    baseline_dims = ['baseline', 'antenna_name']
    for axis in ['x_axis', 'y_axis', 'iter_axis', 'agg_axis']:
        if inputs[axis] in baseline_dims:
            inputs[axis] = baseline_dim

def _check_axis_inputs(inputs):
    ''' Check x_axis, y_axis, vis_axis, and iter_axis inputs. '''
    x_axis = inputs['x_axis']
    y_axis = inputs['y_axis']
    if x_axis == y_axis:
        raise ValueError(f"Invalid parameter values: x_axis {x_axis} cannot be same as y_axis {y_axis}.")

    iter_axis = inputs['iter_axis']
    if iter_axis and iter_axis in (x_axis, y_axis):
        raise ValueError(f"Invalid parameter value: iter_axis {iter_axis} cannot be x_axis ({x_axis}) or y_axis ({y_axis}).")

    data_dims = inputs['data_dims'] if 'data_dims' in inputs else None
    if data_dims:
        if x_axis not in data_dims or y_axis not in data_dims:
            raise ValueError(f"Invalid parameter value: x and y axis must be a data dimension in {data_dims}.")
        if iter_axis and iter_axis not in data_dims:
            raise ValueError(f"Invalid parameter value: iter_axis {iter_axis} must be a data dimension in {data_dims}.")

    vis_axis = inputs['vis_axis']
    if vis_axis not in VIS_AXIS_OPTIONS:
        raise ValueError(f"Invalid parameter value: vis_axis {vis_axis} must be one of {VIS_AXIS_OPTIONS}")

def _check_agg_inputs(inputs):
    ''' Check aggregator and agg_axis. Set agg_axis if not set. '''
    aggregator = inputs['aggregator']
    agg_axis = inputs['agg_axis']

    x_axis = inputs['x_axis']
    y_axis = inputs['y_axis']
    data_dims = inputs['data_dims'] if 'data_dims' in inputs else None

    if aggregator and aggregator not in AGGREGATOR_OPTIONS:
        raise ValueError(f"Invalid parameter value: aggregator {aggregator} must be None or one of {AGGREGATOR_OPTIONS}.")

    if agg_axis:
        if not isinstance(agg_axis, str) and not isinstance(agg_axis, list):
            raise TypeError(f"Invalid parameter type: agg_axis {agg_axis} must be str or list.")

        # Make agg_axis a list
        if isinstance(agg_axis, str):
            agg_axis = [agg_axis]

        for axis in agg_axis:
            if axis in (x_axis, y_axis):
                raise ValueError(f"Invalid parameter value: agg_axis {agg_axis} cannot be x_axis ({x_axis}) or y_axis ({y_axis}).")
            if data_dims and axis not in data_dims:
                raise ValueError(f"Invalid parameter value: agg_axis {axis} must be a data dimension in {data_dims}.")
    elif aggregator and data_dims:
        # Set agg_axis to non-plotted dim axes
        agg_axis = data_dims.copy()
        agg_axis.remove(x_axis)
        agg_axis.remove(y_axis)
        if 'iter_axis' in inputs and inputs['iter_axis']:
            agg_axis.remove(inputs['iter_axis'])
    inputs['agg_axis'] = agg_axis

def _check_color_inputs(inputs):
    if inputs['color_mode']:
        color_mode = inputs['color_mode'].lower()
        valid_color_modes = ['auto', 'manual']
        if color_mode not  in valid_color_modes:
            raise ValueError(f"Invalid parameter value: color_mode {color_mode} must be None or one of {valid_color_modes}.")
        inputs['color_mode'] = color_mode

    if inputs['color_range']:
        if not (isinstance(inputs['color_range'], tuple) and len(inputs['color_range']) == 2):
            raise ValueError("Invalid parameter type: color_range must be None or a tuple of (min, max).")

def _check_other_inputs(inputs):
    if inputs['iter_range']:
        if not (isinstance(inputs['iter_range'], tuple) and len(inputs['iter_range']) == 2):
            raise ValueError("Invalid parameter type: iter_range must be None or a tuple of (start, end).")

    if inputs['subplots']:
        if not (isinstance(inputs['subplots'], tuple) and len(inputs['subplots']) == 2):
            raise ValueError("Invalid parameter type: subplots must be None or a tuple of (rows, columns).")

    if inputs['title'] and not isinstance(inputs['title'], str):
        raise TypeError("Invalid parameter type: title must be None or a string.")



'''
    Create interactive GUI for ms raster plotting
'''

import holoviews as hv
import panel as pn
from vidavis.plot.ms_plot._ms_plot_selectors import (file_selector, title_selector, style_selector,
    axis_selector, aggregation_selector, iteration_selector, selection_selector, plot_starter)

def create_raster_gui(callbacks, data_dims, x_axis, y_axis):
    ''' Use Holoviz Panel to create a dashboard for plot inputs and raster plot display. '''
    # Accordion of widgets for plot inputs
    selectors = get_plot_input_selectors(callbacks, data_dims, x_axis, y_axis)

    # Plot button and spinner while plotting
    init_plot = plot_starter(callbacks['plot_updating'])

    # Dynamic map for plot, with callback when inputs change or location needed
    dmap, points = get_plot_dmap(callbacks, selectors, init_plot)

    return pn.Row(
        pn.Tabs(             # Row [0]
            ('Plot',                                 # Tabs[0]
                pn.Column(
                    dmap * points,  # [0] plot with hv.Points overlay for point_draw
                    pn.WidgetBox(), # [1] cursor location
                )
            ),
            ('Plot Inputs', pn.Column()),            # Tabs[1]
            ('Locate Selected Points', pn.Column()), # Tabs[2]
            ('Locate Selected Box', pn.Column()),    # Tabs[3]
            sizing_mode='stretch_width',
        ),
        pn.Spacer(width=10), # Row [1]
        pn.Column(  # Row [2]
            pn.Spacer(height=25), # Column[0]
            selectors,            # Column[1]
            init_plot,            # Column[2]
            width_policy='min',
            width=400,
            sizing_mode='stretch_height',
        ),
        sizing_mode='stretch_height',
    )

def get_plot_input_selectors(callbacks, data_dims, x_axis, y_axis):
    ''' Create accordion of widgets for plot inputs selection '''
    # Select MS
    file_selectors = file_selector('Path to MeasurementSet (ms or zarr) for plot', '~' , callbacks['filename'])

    # Select style - colormaps, colorbar, color limits
    style_selectors = style_selector(callbacks['style'], callbacks['color'])

    # Select x, y, and vis axis
    axis_selectors = axis_selector(x_axis, y_axis, data_dims, True, callbacks['axes'])

    # Select from ProcessingSet and MeasurementSet
    selection_selectors = selection_selector(callbacks['select_ps'], callbacks['select_ms'])

    # Generic axis options, updated when ms is set
    axis_options = data_dims if data_dims else []

    # Select aggregator and axes to aggregate
    agg_selectors = aggregation_selector(axis_options, callbacks['aggregation'])

    # Select iter_axis and iter value or range
    iter_selectors = iteration_selector(axis_options, callbacks['iter_values'], callbacks['iteration'])

    # Set title
    title_input = title_selector(callbacks['title'])

    # Put user input widgets in accordion with only one card active at a time (toggle)
    selectors = pn.Accordion(
        ("Select file", file_selectors),         # [0]
        ("Plot style", style_selectors),         # [1]
        ("Data Selection", selection_selectors), # [2]
        ("Plot axes", axis_selectors),           # [3]
        ("Aggregation", agg_selectors),          # [4]
        ("Iteration", iter_selectors),           # [5]
        ("Plot title", title_input),             # [6]
    )
    selectors.toggle = True
    return selectors

def get_plot_dmap(callbacks, selectors, init_plot):
    ''' Dynamic map for updating plot from callback function '''
    # Connect plot to filename and plot button; add streams for cursor position, drawn points, and selected box
    # 'update_plot' callback must have parameters (ms, do_plot, x, y, data, bounds)
    points = hv.Points([]).opts(
        size=5,
        fill_color='white'
    )
    dmap = hv.DynamicMap(
        pn.bind(
            callbacks['update_plot'],
            ms=selectors[0][0][0],
            do_plot=init_plot[0],
        ),
        streams=[
            hv.streams.PointerXY(),              # cursor location (x, y)
            hv.streams.PointDraw(source=points), # fixed cursor location (data)
            hv.streams.BoundsXY()                # box location (bounds)
        ]
    )
    return dmap, points



"""
Functions for reading MSv2 or zarr MSv4 into xradio ProcessingSet
"""

import os.path

from xradio.measurement_set.open_processing_set import open_processing_set

try:
    # requires python-casacore
    from xradio.measurement_set.convert_msv2_to_processing_set import convert_msv2_to_processing_set
    __HAVE_CASACORE = True
except ImportError:
    __HAVE_CASACORE = False

def get_processing_set(ms_path, logger):
    '''
    Read msv2 or zarr file into processing set

    Args:
        ms_path (str): path to MSv2 or MSv4 zarr file
    Returns:
        xradio ProcessingSet
    '''

    if not os.path.exists(ms_path):
        raise RuntimeError(f"Visibility file {ms_path} does not exist")

    # Remove trailing / or entire path is basename with no extension
    if ms_path[-1] == '/':
        ms_path = ms_path[:-1]

    basename, ext = os.path.splitext(ms_path)
    if ext == ".zarr":
        zarr_path = ms_path
    else:
        zarr_path = basename + ".ps.zarr"
        if not os.path.exists(zarr_path):
            if not __HAVE_CASACORE:
                raise RuntimeError("Cannot convert MSv2 to xradio zarr file: python-casacore not installed.")
            logger.info(f"Converting input MS {ms_path} to zarr {zarr_path}")
            convert_msv2_to_processing_set(
                in_file=ms_path,
                out_file=zarr_path
                )

    if not os.path.exists(zarr_path):
        raise RuntimeError("Zarr file does not exist")

    ps = open_processing_set(zarr_path)
    if not ps or len(ps) == 0:
        raise RuntimeError("Failed to read measurement set into processing set.")
    logger.info(f"Processing set contains {len(ps)} msv4 datasets.")

    return ps, zarr_path
